package org.xbib.jdbc.query;

import java.math.BigDecimal;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Entry point for getting configuration parameters. This isn't intended as
 * a be-all, end-all configuration solution. Just a way of easily specifying
 * multiple read-only sources for configuration with a nice fluent syntax.
 */
public interface Config extends Function<String, String>, Supplier<Config> {
    /**
     * Convenience method for fluent syntax.
     *
     * @return a builder for specifying from where configuration should be loaded
     */
    static ConfigFrom from() {
        return new ConfigFromImpl();
    }

    // TODO add: String originalKey(String key) to find out the key before prefixing or other manipulation

    /**
     * @return a trimmed, non-empty string, or null
     */
    String getString(String key);

    /**
     * @return a trimmed, non-empty string
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    String getStringOrThrow(String key);

    String getString(String key, String defaultValue);

    /**
     * Same as {@link #getString(String)}. Useful for passing configs around
     * without static dependencies.
     */
    @Override
    default String apply(String key) {
        return getString(key);
    }

    @Override
    default Config get() {
        return this;
    }

    
    Integer getInteger(String key);

    int getInteger(String key, int defaultValue);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    int getIntegerOrThrow(String key);

    
    Long getLong(String key);

    long getLong(String key, long defaultValue);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    long getLongOrThrow(String key);

    
    Float getFloat(String key);

    float getFloat(String key, float defaultValue);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    float getFloatOrThrow(String key);

    
    Double getDouble(String key);

    double getDouble(String key, double defaultValue);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    double getDoubleOrThrow(String key);

    
    BigDecimal getBigDecimal(String key);

    
    BigDecimal getBigDecimal(String key, BigDecimal defaultValue);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    
    BigDecimal getBigDecimalOrThrow(String key);

    /**
     * Read a boolean value from the configuration. The value is not case-sensitivie,
     * and may be either true/false or yes/no. If no value was provided or an invalid
     * value is provided, false will be returned.
     */
    boolean getBooleanOrFalse(String key);

    /**
     * Read a boolean value from the configuration. The value is not case-sensitivie,
     * and may be either true/false or yes/no. If no value was provided or an invalid
     * value is provided, true will be returned.
     */
    boolean getBooleanOrTrue(String key);

    /**
     * @throws ConfigMissingException if no value could be read for the specified key
     */
    boolean getBooleanOrThrow(String key);

    /**
     * Show where configuration is coming from. This is useful to drop in your logs
     * for troubleshooting.
     */
    String sources();
}
