package org.xbib.jdbc.query;

import java.math.BigDecimal;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class handles all of the type conversions, default values, etc.
 * to get from simple string key/value pairs to richer configuration.
 */
public class ConfigImpl implements Config {
    
    private static final Logger log = Logger.getLogger(ConfigFromImpl.class.getName());
    
    private final Function<String, String> provider;
    
    private final String sources;
    
    private final Set<String> failedKeys = new HashSet<>();

    public ConfigImpl(Function<String, String> provider, String sources) {
        this.provider = provider;
        this.sources = sources;
    }

    @Override
    public String getString(String key) {
        return cleanString(key);
    }

   
    @Override
    public String getStringOrThrow(String key) {
        return nonnull(key, getString(key));
    }

   
    @Override
    public String getString(String key, String defaultValue) {
        String stringValue = cleanString(key);
        if (stringValue != null) {
            return stringValue;
        }
        // Make sure the default value is tidied the same way a value would be
        defaultValue = defaultValue.trim();
        if (defaultValue.length() == 0) {
            throw new IllegalArgumentException("Your default value is empty or just whitespace");
        }
        return defaultValue;
    }

    @Override
    public Integer getInteger(String key) {
        String stringValue = cleanString(key);
        try {
            return stringValue == null ? null : Integer.parseInt(stringValue);
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
            return null;
        }
    }

    @Override
    public int getInteger(String key, int defaultValue) {
        Integer value = getInteger(key);
        return value == null ? defaultValue : value;
    }

    @Override
    public int getIntegerOrThrow(String key) {
        return nonnull(key, getInteger(key));
    }

   
    @Override
    public Long getLong(String key) {
        String stringValue = cleanString(key);
        try {
            return stringValue == null ? null : Long.parseLong(stringValue);
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
            return null;
        }
    }

    @Override
    public long getLong(String key, long defaultValue) {
        Long value = getLong(key);
        return value == null ? defaultValue : value;
    }

    @Override
    public long getLongOrThrow(String key) {
        return nonnull(key, getLong(key));
    }

   
    @Override
    public Float getFloat(String key) {
        String stringValue = cleanString(key);
        try {
            return stringValue == null ? null : Float.parseFloat(stringValue);
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
            return null;
        }
    }

    @Override
    public float getFloat(String key, float defaultValue) {
        Float value = getFloat(key);
        return value == null ? defaultValue : value;
    }

    @Override
    public float getFloatOrThrow(String key) {
        return nonnull(key, getFloat(key));
    }

   
    @Override
    public Double getDouble(String key) {
        String stringValue = cleanString(key);
        try {
            return stringValue == null ? null : Double.parseDouble(stringValue);
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
            return null;
        }
    }

    @Override
    public double getDouble(String key, double defaultValue) {
        Double value = getDouble(key);
        return value == null ? defaultValue : value;
    }

    @Override
    public double getDoubleOrThrow(String key) {
        return nonnull(key, getDouble(key));
    }

   
    @Override
    public BigDecimal getBigDecimal(String key) {
        String stringValue = cleanString(key);
        try {
            return stringValue == null ? null : new BigDecimal(stringValue);
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
            return null;
        }
    }

   
    @Override
    public BigDecimal getBigDecimal(String key, BigDecimal defaultValue) {
        BigDecimal value = getBigDecimal(key);
        return value == null ? defaultValue : value;
    }

   
    @Override
    public BigDecimal getBigDecimalOrThrow(String key) {
        return nonnull(key, getBigDecimal(key));
    }

    @Override
    public boolean getBooleanOrFalse(String key) {
        return parseBoolean(cleanString(key), false);
    }

    @Override
    public boolean getBooleanOrTrue(String key) {
        return parseBoolean(cleanString(key), true);
    }

    @Override
    public boolean getBooleanOrThrow(String key) {
        String value = nonnull(key, cleanString(key));
        value = value.toLowerCase();
        if (value.equals("yes") || value.equals("true")) {
            return true;
        }
        if (value.equals("no") || value.equals("false")) {
            return false;
        }
        throw new ConfigMissingException("Unrecognized boolean value for config key: " + key);
    }

    @Override
    public String sources() {
        return sources;
    }

    private <T> T nonnull(String key, T value) {
        if (value == null) {
            throw new ConfigMissingException("No value for config key: " + key);
        }
        return value;
    }

    private boolean parseBoolean(String value, boolean defaultValue) {
        if (value != null) {
            value = value.toLowerCase();
            if (value.equals("yes") || value.equals("true")) {
                return true;
            }
            if (value.equals("no") || value.equals("false")) {
                return false;
            }
        }
        return defaultValue;
    }

    private String cleanString(String key) {
        String value = null;
        try {
            value = provider.apply(key);
            if (value != null) {
                value = value.trim();
                if (value.length() == 0) {
                    value = null;
                }
            }
        } catch (Exception e) {
            if (!failedKeys.contains(key)) {
                log.log(Level.SEVERE, "Could not load config value for key (this message will only be logged once): " + key, e);
                failedKeys.add(key);
            }
        }
        return value;
    }
}
