package org.xbib.jdbc.query;

import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Primary class for accessing a relational (SQL) database.
 */
public class DatabaseImpl implements Database {
    
    private static final Logger log = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;

    private final Options options;

    public DatabaseImpl(Flavor flavor) {
        this(new OptionsDefault(flavor));
    }

    public DatabaseImpl(Options options) {
        this(null, options);
    }

    public DatabaseImpl(Connection connection, Options options) {
        this.connection = connection;
        this.options = options;
    }

    @Override
    public DatabaseImpl get() {
        return this;
    }

    @Override
    public SqlInsert toInsert(String sql) {
        return new SqlInsertImpl(connection, sql, options);
    }

    @Override
    public SqlInsert toInsert(Sql sql) {
        return new SqlInsertImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlSelect toSelect(String sql) {
        return new SqlSelectImpl(connection, sql, options);
    }

    @Override
    public SqlSelect toSelect(Sql sql) {
        return new SqlSelectImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlUpdate toUpdate(String sql) {
        return new SqlUpdateImpl(connection, sql, options);
    }

    @Override
    public SqlUpdate toUpdate(Sql sql) {
        return new SqlUpdateImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlUpdate toDelete(String sql) {
        return new SqlUpdateImpl(connection, sql, options);
    }

    @Override
    public SqlUpdate toDelete(Sql sql) {
        return new SqlUpdateImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public Ddl ddl(String sql) {
        return new DdlImpl(connection, sql, options);
    }

    @Override
    public Long nextSequenceValue(/*@Untainted*/ String sequenceName) {
        return toSelect(flavor().sequenceSelectNextVal(sequenceName)).queryLongOrNull();
    }

    @Override
    public LocalDateTime nowPerApp() {
        return options.currentDate();
    }

    public void commitNow() {
        if (options.ignoreTransactionControl()) {
            log.fine("Ignoring call to commitNow()");
            return;
        }
        if (!options.allowTransactionControl()) {
            throw new DatabaseException("Calls to commitNow() are not allowed");
        }
        try {
            connection.commit();
        } catch (Exception e) {
            throw new DatabaseException("Unable to commit transaction", e);
        }
    }

    public void rollbackNow() {
        if (options.ignoreTransactionControl()) {
            log.fine("Ignoring call to rollbackNow()");
            return;
        }
        if (!options.allowTransactionControl()) {
            throw new DatabaseException("Calls to rollbackNow() are not allowed");
        }
        try {
            connection.rollback();
        } catch (Exception e) {
            throw new DatabaseException("Unable to rollback transaction", e);
        }
    }
    
    @Override
    public Connection underlyingConnection() {
        if (!options.allowConnectionAccess()) {
            throw new DatabaseException("Calls to underlyingConnection() are not allowed");
        }

        return connection;
    }
    
    @Override
    public Options options() {
        return options;
    }
    
    @Override
    public Flavor flavor() {
        return options.flavor();
    }

    
    @Override
    public When when() {
        return new When(options.flavor());
    }

    @Override
    public void dropSequenceQuietly(/*@Untainted*/ String sequenceName) {
        ddl(flavor().sequenceDrop(sequenceName)).executeQuietly();
    }

    @Override
    public void dropTableQuietly(/*@Untainted*/ String tableName) {
        if (flavor() == Flavor.postgresql || flavor() == Flavor.hsqldb) {
            ddl("drop table if exists " + tableName).executeQuietly();
        } else {
            ddl("drop table " + tableName).executeQuietly();
        }
    }

    @Override
    public boolean tableExists(String tableName) throws DatabaseException {

        String schemaName = null;
        Method getSchema = null;

        try {
            // Use reflections to see if connection.getSchema API exists. It should exist for any JDBC7 or later implementation
            // We still support Oracle 11 with odbc6, however, so we can't assume it's there.
            getSchema = connection.getClass().getDeclaredMethod("getSchema");
        } catch (NoSuchMethodException noMethodExc) {
            // Expected if method does not exist - just let it go
        }

        try {
            if (getSchema != null) {
                schemaName = ((String) getSchema.invoke(connection, new Object[0]));
            } else if (flavor() == Flavor.oracle) {
                // Oracle defaults to user name schema - use that.
                log.warning("Connection getSchema API was not found.  Defaulting to Oracle user name schema." +
                        "If this is not appropriate, please use tableExists(tableName, schemaName) API or upgrade to ojdbc7 or later");
                schemaName = connection.getMetaData().getUserName();
            }
            if (schemaName == null) {
                // connection.getSchema API was supported starting at JDK1.7.  Method should not be null.
                throw new NullPointerException("Unable to retrieve schema name.");
            }

        } catch (Exception exc) {
            throw new DatabaseException("Unable to determine the schema. " +
                    "Please use tableExists(tableName, schemaName API) or upgrade to a JDBC7 driver or later.", exc);
        }

        return tableExists(tableName, schemaName);
    }

    @Override
    public boolean tableExists(String tableName, String schemaName) throws DatabaseException {
        if (tableName != null && connection != null) {
            try {
                DatabaseMetaData metadata = connection.getMetaData();
                String normalizedTable = normalizeTableName(tableName);
                ResultSet resultSet = metadata.getTables(connection.getCatalog(), schemaName, normalizedTable, new String[]{"TABLE", "VIEW"});
                while (resultSet.next()) {
                    if (normalizedTable.equals(resultSet.getString("TABLE_NAME"))) {
                        return true;
                    }
                }
            } catch (SQLException exc) {
                throw new DatabaseException("Unable to look up table " + tableName
                        + " in schema  " + schemaName + " : " + exc.getMessage(),
                        exc);
            }
        }
        return false;
    }

    @Override
    public Map<String, Integer> getColumnSizesOfTable(String tableName) {
        Map<String, Integer> map = new HashMap<>();
        if (tableName != null && connection != null) {
            try {
                DatabaseMetaData metaData = connection.getMetaData();
                String normalizedTable = normalizeTableName(tableName);
                ResultSet rs = metaData.getColumns(null, null, normalizedTable, "%");
                ResultSetMetaData rsmd = rs.getMetaData();
                int cols = rsmd.getColumnCount();
                while (rs.next()) {
                    String name = "";
                    Integer size = 0;
                    for (int i = 1; i <= cols; i++) {
                        String label = rsmd.getColumnName(i);
                        Object v = rs.getObject(i);
                        if (label.equals("COLUMN_NAME")) {
                            name = (String) v;
                        }
                        if (label.equals("COLUMN_SIZE")) {
                            size = (Integer) v;
                        }
                    }
                    map.put(name, size);
                }
                rs.close();
            } catch (SQLException exc) {
                throw new DatabaseException("Unable to look up table " + tableName + " : " + exc.getMessage(), exc);
            }
        }
        return map;
    }

    @Override
    public String normalizeTableName(String tableName) {
        if (tableName == null) {
            return tableName;
        }

        // If user gave us a quoted string, leave it alone for look up
        if (tableName.length() > 2) {
            if (tableName.startsWith("\"") && tableName.endsWith("\"")) {
                // Remove quotes and return as is.
                return tableName.substring(1, tableName.length() - 1);
            }
        }

        if (flavor().isNormalizedUpperCase()) {
            return tableName.toUpperCase();
        }

        return tableName.toLowerCase();
    }

    @Override
    public void assertTimeSynchronized(long millisToWarn, long millisToError) {
        toSelect("select ?" + flavor().fromAny())
                .argDateNowPerDb().queryFirstOrNull(r -> {
                    LocalDateTime appDate = nowPerApp();
                    LocalDateTime dbDate = r.getLocalDateTimeOrNull();
                    if (dbDate == null) {
                        throw new DatabaseException("Expecting a date in the result");
                    }
                    Duration duration = Duration.between(appDate, dbDate).abs();
                    if (duration.getSeconds() > 3600) {
                        throw new DatabaseException("App and db time are over an hour apart (check your timezones) app: "
                                + DateTimeFormatter.ISO_INSTANT.format(appDate) + " db: "
                                + DateTimeFormatter.ISO_INSTANT.format(dbDate));
                    }
                    if (duration.getSeconds() * 1000 > millisToError) {
                        throw new DatabaseException("App and db time over " + millisToError + " millis apart (check your clocks) app: "
                                + DateTimeFormatter.ISO_INSTANT.format(appDate) + " db: "
                                + DateTimeFormatter.ISO_INSTANT.format(dbDate));
                    }
                    if (duration.getSeconds() * 1000 > millisToWarn) {
                        log.warning("App and db time are over " + millisToWarn + " millis apart (check your clocks) app: "
                                + DateTimeFormatter.ISO_INSTANT.format(appDate) + " db: "
                                + DateTimeFormatter.ISO_INSTANT.format(dbDate));
                    }
                    return null;
                });
    }

    @Override
    public void assertTimeSynchronized() {
        assertTimeSynchronized(10000, 30000);
    }
}
