package org.xbib.jdbc.query;

import org.xbib.jdbc.query.util.DebugSql;
import org.xbib.jdbc.query.util.Metric;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;

public class DdlImpl implements Ddl {

    private static final Logger log = Logger.getLogger(Database.class.getName());

    private static final Logger logQuiet = Logger.getLogger(Database.class.getName() + ".quiet");

    private final Connection connection;

    private final String sql;

    private final Options options;

    DdlImpl(Connection connection, String sql, Options options) {
        this.connection = connection;
        this.sql = sql;
        this.options = options;
    }

    private void updateInternal(boolean quiet) {
        CallableStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            ps = connection.prepareCall(sql);

            metric.checkpoint("prep");
            ps.execute();
            metric.checkpoint("exec");
            isSuccess = true;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(sql, null, errorCode, options), e);
        } finally {
            close(ps);
            metric.checkpoint("close");
            // PostgreSQL requires explicit commit since we are running with setAutoCommit(false)
            commit(connection);
            metric.done("commit");
            if (isSuccess) {
                DebugSql.logSuccess("DDL", log, metric, sql, null, options);
            } else if (quiet) {
                DebugSql.logWarning("DDL", logQuiet, metric, errorCode, sql, null, options, logEx);
            } else {
                DebugSql.logError("DDL", log, metric, errorCode, sql, null, options, logEx);
            }
        }
    }

    @Override
    public void execute() {
        updateInternal(false);
    }

    @Override
    public void executeQuietly() {
        try {
            updateInternal(true);
        } catch (DatabaseException e) {
            // Ignore, as requested
        }
    }

    private void close(Statement s) {
        if (s != null) {
            try {
                s.close();
            } catch (Exception e) {
                log.log(Level.SEVERE, "Caught exception closing the Statement", e);
            }
        }
    }

    private void commit(Connection c) {
        if (c != null) {
            try {
                c.commit();
            } catch (Exception e) {
                log.log(Level.SEVERE, "Caught exception on commit", e);
            }
        }
    }
}
