package org.xbib.jdbc.query;

import org.xbib.jdbc.query.util.RewriteArg;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Convenience class to allow use of (:mylabel) for SQL parameters in addition to
 * positional (?) parameters. This doesn't do any smart parsing of the SQL, it is just
 * looking for ':' and '?' characters. If the SQL needs to include an actual ':' or '?'
 * character, use two of them ('::' or '??'), and they will be replaced with a
 * single ':' or '?'.
 */
public class MixedParameterSql {

    private final String sqlToExecute;

    private final Object[] args;

    public MixedParameterSql(String sql, List<Object> positionalArgs, Map<String, Object> nameToArg) {
        if (positionalArgs == null) {
            positionalArgs = new ArrayList<>();
        }
        if (nameToArg == null) {
            nameToArg = new HashMap<>();
        }

        StringBuilder newSql = new StringBuilder(sql.length());
        List<String> argNamesList = new ArrayList<>();
        List<String> rewrittenArgs = new ArrayList<>();
        List<Object> argsList = new ArrayList<>();
        int searchIndex = 0;
        int currentPositionalArg = 0;
        while (searchIndex < sql.length()) {
            int nextColonIndex = sql.indexOf(':', searchIndex);
            int nextQmIndex = sql.indexOf('?', searchIndex);

            if (nextColonIndex < 0 && nextQmIndex < 0) {
                newSql.append(sql.substring(searchIndex));
                break;
            }

            if (nextColonIndex >= 0 && (nextQmIndex == -1 || nextColonIndex < nextQmIndex)) {
                // The next parameter we found is a named parameter (":foo")
                if (nextColonIndex > sql.length() - 2) {
                    // Probably illegal sql, but handle boundary condition
                    break;
                }

                // Allow :: as escape for :
                if (sql.charAt(nextColonIndex + 1) == ':') {
                    newSql.append(sql, searchIndex, nextColonIndex + 1);
                    searchIndex = nextColonIndex + 2;
                    continue;
                }

                int endOfNameIndex = nextColonIndex + 1;
                while (endOfNameIndex < sql.length() && Character.isJavaIdentifierPart(sql.charAt(endOfNameIndex))) {
                    endOfNameIndex++;
                }
                newSql.append(sql, searchIndex, nextColonIndex);
                String paramName = sql.substring(nextColonIndex + 1, endOfNameIndex);
                if (!"".equals(paramName)) {
                    boolean secretParam = paramName.startsWith("secret");
                    Object arg = nameToArg.get(paramName);
                    if (arg instanceof RewriteArg) {
                        newSql.append(((RewriteArg) arg).getSql());
                        rewrittenArgs.add(paramName);
                    } else {
                        newSql.append('?');
                        if (nameToArg.containsKey(paramName)) {
                            argsList.add(secretParam ? new SecretArg(arg) : arg);
                        } else {
                            throw new DatabaseException("The SQL requires parameter ':" + paramName + "' but no value was provided");
                        }
                        argNamesList.add(paramName);
                    }
                }
                searchIndex = endOfNameIndex;
            } else {
                // The next parameter we found is a positional parameter ("?")

                // Allow ?? as escape for ?
                if (nextQmIndex < sql.length() - 1 && sql.charAt(nextQmIndex + 1) == '?') {
                    newSql.append(sql, searchIndex, nextQmIndex + 1);
                    searchIndex = nextQmIndex + 2;
                    continue;
                }

                newSql.append(sql, searchIndex, nextQmIndex);
                if (currentPositionalArg >= positionalArgs.size()) {
                    throw new DatabaseException("Not enough positional parameters (" + positionalArgs.size() + ") were provided");
                }
                if (positionalArgs.get(currentPositionalArg) instanceof RewriteArg) {
                    newSql.append(((RewriteArg) positionalArgs.get(currentPositionalArg)).getSql());
                } else {
                    newSql.append('?');
                    argsList.add(positionalArgs.get(currentPositionalArg));
                }
                currentPositionalArg++;
                searchIndex = nextQmIndex + 1;
            }
        }
        this.sqlToExecute = newSql.toString();
        args = argsList.toArray(new Object[argsList.size()]);

        // Sanity check number of arguments to provide a better error message
        if (currentPositionalArg != positionalArgs.size()) {
            throw new DatabaseException("Wrong number of positional parameters were provided (expected: "
                    + currentPositionalArg + ", actual: " + positionalArgs.size() + ")");
        }
        if (nameToArg.size() > args.length - Math.max(0, positionalArgs.size() - 1) + rewrittenArgs.size()) {
            Set<String> unusedNames = new HashSet<>(nameToArg.keySet());
            unusedNames.removeAll(argNamesList);
            unusedNames.removeAll(rewrittenArgs);
            if (!unusedNames.isEmpty()) {
                throw new DatabaseException("These named parameters do not exist in the query: " + unusedNames);
            }
        }
    }

    public String getSqlToExecute() {
        return sqlToExecute;
    }

    public Object[] getArgs() {
        return args;
    }

}
