package org.xbib.jdbc.query;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.ResultSetMetaData;
import java.time.LocalDate;
import java.time.LocalDateTime;

/**
 * Interface for reading results from a database query.
 */
public interface Row {
    /**
     * Obtain the names of the columns in the database. You probably want to
     * avoid this method if possible, as the way column names are handled varies
     * by database and driver. For example, Derby and Oracle normally convert
     * column names to uppercase, while PostgreSQL normally converts column
     * names to lowercase. If you do use this method, you might want to either
     * call toUppercase()/toLowercase() or ensure the SQL explicitly specifies
     * parameters with AS "FOO" (including quotes) to ensure your desired name
     * will be honored.
     */
    String[] getColumnLabels();

    /**
     * Get raw access to the underlying JDBC metadata.
     */
    ResultSetMetaData getMetadata();

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to {@code Boolean} or {@code null} as appropriate.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * <p>If you call one of the methods using an explicit column index or column name before
     * calling this method, it will pick up at the next column following the explicit one.
     * For example:</p>
     *
     * <pre>
     * getX();  // column 1
     * getX(5); // or getX("foo") if foo is column 5
     * getX();  // column 6
     * </pre>
     *
     * @return true if the value was "Y", false if it was "N", or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    Boolean getBooleanOrNull();

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to {@code Boolean} or {@code null} as appropriate.
     *
     * @param columnOneBased column number to read (1 is the first column)
     * @return true if the value was "Y", false if it was "N", or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    Boolean getBooleanOrNull(int columnOneBased);

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to {@code Boolean} or {@code null} as appropriate.
     *
     * @param columnName SQL alias of the column to read (use all lowercase)
     * @return true if the value was "Y", false if it was "N", or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    Boolean getBooleanOrNull(String columnName);

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code false}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * <p>If you call one of the methods using an explicit column index or column name before
     * calling this method, it will pick up at the next column following the explicit one.
     * For example:</p>
     *
     * <pre>
     * getX();  // column 1
     * getX(5); // or getX("foo") if foo is column 5
     * getX();  // column 6
     * </pre>
     *
     * @return true if the value was "Y", false if it was either "N" or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrFalse();

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code false}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * @param columnOneBased column number to read (1 is the first column)
     * @return true if the value was "Y", false if it was either "N" or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrFalse(int columnOneBased);

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code false}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * @param columnName SQL alias of the column to read (use all lowercase)
     * @return true if the value was "Y", false if it was either "N" or null
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrFalse(String columnName);

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code true}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * <p>If you call one of the methods using an explicit column index or column name before
     * calling this method, it will pick up at the next column following the explicit one.
     * For example:</p>
     *
     * <pre>
     * getX();  // column 1
     * getX(5); // or getX("foo") if foo is column 5
     * getX();  // column 6
     * </pre>
     *
     * @return true if the value was either "Y" or null, false if it was "N"
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrTrue();

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code true}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * @param columnOneBased column number to read (1 is the first column)
     * @return true if the value was either "Y" or null, false if it was "N"
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrTrue(int columnOneBased);

    /**
     * Boolean values are represented as strings {@code "Y"} or {@code "N"} in the database,
     * typically in a {@code CHAR(1)} column. This reads the value and converts it
     * to a {@code boolean}. If the value is {@code null}, it will be converted to {@code true}.
     *
     * <p>This is a short-hand method that reads columns in order, starting
     * with the first, and automatically incrementing the column index.</p>
     *
     * @param columnName SQL alias of the column to read (use all lowercase)
     * @return true if the value was either "Y" or null, false if it was "N"
     * @throws DatabaseException if the value was something other than Y, N, or null
     */
    boolean getBooleanOrTrue(String columnName);
    
    Integer getIntegerOrNull();
    
    Integer getIntegerOrNull(int columnOneBased);
    
    Integer getIntegerOrNull(String columnName);

    int getIntegerOrZero();

    int getIntegerOrZero(int columnOneBased);

    int getIntegerOrZero(String columnName);
    
    Long getLongOrNull();

    Long getLongOrNull(int columnOneBased);
    
    Long getLongOrNull(String columnName);

    long getLongOrZero();

    long getLongOrZero(int columnOneBased);

    long getLongOrZero(String columnName);
    
    Float getFloatOrNull();

    Float getFloatOrNull(int columnOneBased);

    Float getFloatOrNull(String columnName);

    float getFloatOrZero();

    float getFloatOrZero(int columnOneBased);

    float getFloatOrZero(String columnName);

    Double getDoubleOrNull();

    Double getDoubleOrNull(int columnOneBased);

    Double getDoubleOrNull(String columnName);

    double getDoubleOrZero();

    double getDoubleOrZero(int columnOneBased);

    double getDoubleOrZero(String columnName);

    /**
     * Note this method attempts to correct for "artifical" scale due to the database
     * representation. Some databases will pad the number out to "full precision". This
     * method tries to reduce scale if there is zero padding to the right of the decimal.
     */
    
    BigDecimal getBigDecimalOrNull();

    
    BigDecimal getBigDecimalOrNull(int columnOneBased);

    
    BigDecimal getBigDecimalOrNull(String columnName);

   
    BigDecimal getBigDecimalOrZero();

   
    BigDecimal getBigDecimalOrZero(int columnOneBased);

   
    BigDecimal getBigDecimalOrZero(String columnName);

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getStringOrNull();

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getStringOrNull(int columnOneBased);

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getStringOrNull(String columnName);

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getStringOrEmpty();

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getStringOrEmpty(int columnOneBased);

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getStringOrEmpty(String columnName);

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getClobStringOrNull();

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getClobStringOrNull(int columnOneBased);

    /**
     * @return the value, or null if it is SQL null; never returns the empty string
     */
    
    String getClobStringOrNull(String columnName);

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getClobStringOrEmpty();

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getClobStringOrEmpty(int columnOneBased);

    /**
     * @return the value, or the empty string if it is SQL null; never returns null
     */
   
    String getClobStringOrEmpty(String columnName);

    /**
     * @return the value, or null if it is SQL null
     */
    
    Reader getClobReaderOrNull();

    /**
     * @return the value, or null if it is SQL null
     */
    
    Reader getClobReaderOrNull(int columnOneBased);

    /**
     * @return the value, or null if it is SQL null
     */
    
    Reader getClobReaderOrNull(String columnName);

    /**
     * @return the value, or a StringReader containing the empty string if it is SQL null
     */
   
    Reader getClobReaderOrEmpty();

    /**
     * @return the value, or a StringReader containing the empty string if it is SQL null
     */
   
    Reader getClobReaderOrEmpty(int columnOneBased);

    /**
     * @return the value, or a StringReader containing the empty string if it is SQL null
     */
   
    Reader getClobReaderOrEmpty(String columnName);

    
    byte[] getBlobBytesOrNull();

    
    byte[] getBlobBytesOrNull(int columnOneBased);

    
    byte[] getBlobBytesOrNull(String columnName);

   
    byte[] getBlobBytesOrZeroLen();

   
    byte[] getBlobBytesOrZeroLen(int columnOneBased);

   
    byte[] getBlobBytesOrZeroLen(String columnName);

    
    InputStream getBlobInputStreamOrNull();

    
    InputStream getBlobInputStreamOrNull(int columnOneBased);

    
    InputStream getBlobInputStreamOrNull(String columnName);

   
    InputStream getBlobInputStreamOrEmpty();

   
    InputStream getBlobInputStreamOrEmpty(int columnOneBased);
   
    InputStream getBlobInputStreamOrEmpty(String columnName);

    LocalDateTime getLocalDateTimeOrNull();
    
    LocalDateTime getLocalDateTimeOrNull(int columnOneBased);
    
    LocalDateTime getLocalDateTimeOrNull(String columnName);

    /**
     * Retrieve column as LocalDate, .i.e, date with no time.
     *
     * @return LocalDate of the database column value
     */
    LocalDate getLocalDateOrNull();

    /**
     * Get the Date field, with no timestamp
     *
     * @param columnOneBased column number starting at 1, not 0
     * @return LocalDate of the column value
     */
    
    LocalDate getLocalDateOrNull(int columnOneBased);

    /**
     * Get the Date field, with no timestamp
     *
     * @param columnName column name to retrieve
     * @return LocalDate of the column value
     */
    
    LocalDate getLocalDateOrNull(String columnName);
}
