package org.xbib.jdbc.query;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Stack;

/**
 * This class is useful for dynamically generating SQL. It can "buffer" the
 * various arg*() calls and replay them later via the apply(sqlArgs) methods.
 */
public class Sql implements SqlInsert.Apply, SqlUpdate.Apply, SqlSelect.Apply {
    private final StringBuilder sql = new StringBuilder();
    private final Stack<Boolean> listFirstItem = new Stack<>();
    private List<SqlArgs> batched;
    private SqlArgs sqlArgs = new SqlArgs();

    public Sql() {
    }

    public Sql(/*@Untainted*/ String sql) {
        this.sql.append(sql);
    }

    public static Sql insert(/*@Untainted*/ String table, SqlArgs args) {
        return insert(table, Collections.singletonList(args));
    }

    public static Sql insert(/*@Untainted*/ String table, List<SqlArgs> args) {
        Sql sql = null;
        List<String> expectedColumns = null;

        for (SqlArgs arg : args) {
            if (arg.positionalCount() > 0) {
                throw new DatabaseException("The SqlArgs must all be named to do this");
            }
            List<String> columns = arg.names();
            if (columns.size() < 1) {
                throw new DatabaseException("You must add named arguments to SqlArgs");
            }
            if (sql == null) {
                expectedColumns = columns;
                sql = new Sql("insert into ").append(table);
                sql.listStart(" (");
                for (String column : columns) {
                    sql.listSeparator(",");
                    sql.append(column);
                }
                sql.listEnd(") values (");
                sql.appendQuestionMarks(columns.size());
                sql.append(")");
            } else {
                if (!expectedColumns.equals(columns)) {
                    throw new DatabaseException("The columns for all rows in a batch must match. \nFirst: " + expectedColumns
                            + "\nCurrent: " + columns);
                }
                sql.batch();
            }
            sql.setSqlArgs(arg.makePositional());
        }
        return sql;
    }

    public Sql setSqlArgs(SqlArgs args) {
        sqlArgs = args;
        return this;
    }

    public Sql batch() {
        if (sqlArgs.argCount() > 0) {
            if (batched == null) {
                batched = new ArrayList<>();
            }
            batched.add(sqlArgs);
            sqlArgs = new SqlArgs();
        }
        return this;
    }

    public int argCount() {
        return sqlArgs.argCount();
    }

    public Sql appendQuestionMarks(int howMany) {
        boolean first = true;
        for (int i = 0; i < howMany; i++) {
            if (first) {
                first = false;
                append("?");
            } else {
                append(",?");
            }
        }
        return this;
    }

    public Sql append(/*@Untainted*/ String sql) {
        this.sql.append(sql);
        return this;
    }

    public Sql append(boolean value) {
        this.sql.append(value);
        return this;
    }

    public Sql append(int value) {
        this.sql.append(value);
        return this;
    }

    public Sql append(long value) {
        this.sql.append(value);
        return this;
    }

    public Sql append(float value) {
        this.sql.append(value);
        return this;
    }

    public Sql append(double value) {
        this.sql.append(value);
        return this;
    }

    public Sql deleteCharAt(int index) {
        this.sql.deleteCharAt(index);
        return this;
    }

    public Sql replace(int start, int end, /*@Untainted*/ String str) {
        this.sql.replace(start, end, str);
        return this;
    }

    public Sql insert(int offset, /*@Untainted*/ String str) {
        this.sql.insert(offset, str);
        return this;
    }

    public Sql insert(int offset, boolean value) {
        this.sql.insert(offset, value);
        return this;
    }

    public Sql insert(int offset, int value) {
        this.sql.insert(offset, value);
        return this;
    }

    public Sql insert(int offset, long value) {
        this.sql.insert(offset, value);
        return this;
    }

    public Sql insert(int offset, double value) {
        this.sql.insert(offset, value);
        return this;
    }

    public Sql insert(int offset, float value) {
        this.sql.insert(offset, value);
        return this;
    }

    public int indexOf(String str) {
        return this.sql.indexOf(str);
    }

    public int indexOf(String str, int fromIndex) {
        return this.sql.indexOf(str, fromIndex);
    }

    public int lastIndexOf(String str) {
        return this.sql.lastIndexOf(str);
    }

    public int lastIndexOf(String str, int fromIndex) {
        return this.sql.lastIndexOf(str, fromIndex);
    }

    /**
     * Appends the bit of sql and notes that a list, or a sublist, has started.
     * <p>
     * Each list started must have be ended. "Lists" are only to support using listSeparator(sep)
     */
    public Sql listStart(/*@Untainted*/ String sql) {
        listFirstItem.push(true);
        return append(sql);
    }

    /**
     * Appends the passed bit of sql only if a previous item has already been appended,
     * and notes that the list is not empty.
     */
    public Sql listSeparator(/*@Untainted*/ String sql) {
        if (listFirstItem.peek()) {
            listFirstItem.pop();
            listFirstItem.push(false);
            return this;
        } else {
            return append(sql);
        }
    }


    public Sql listEnd(/*@Untainted*/ String sql) {
        listFirstItem.pop();
        return append(sql);
    }

    /*@Untainted*/
    public String sql() {
        return sql.toString();
    }

    /**
     * Same as sql(), provided for drop-in compatibility with StringBuilder.
     */
    /*@Untainted*/
    public String toString() {
        return sql();
    }
    
    public Sql argBoolean(Boolean arg) {
        sqlArgs.argBoolean(arg);
        return this;
    }

    
    public Sql argBoolean( String argName, Boolean arg) {
        sqlArgs.argBoolean(argName, arg);
        return this;
    }

    
    public Sql argInteger(Integer arg) {
        sqlArgs.argInteger(arg);
        return this;
    }

    
    public Sql argInteger( String argName, Integer arg) {
        sqlArgs.argInteger(argName, arg);
        return this;
    }

    
    public Sql argLong(Long arg) {
        sqlArgs.argLong(arg);
        return this;
    }

    
    public Sql argLong(String argName, Long arg) {
        sqlArgs.argLong(argName, arg);
        return this;
    }

    
    public Sql argFloat(Float arg) {
        sqlArgs.argFloat(arg);
        return this;
    }

    
    public Sql argFloat( String argName, Float arg) {
        sqlArgs.argFloat(argName, arg);
        return this;
    }

    
    public Sql argDouble(Double arg) {
        sqlArgs.argDouble(arg);
        return this;
    }

    
    public Sql argDouble( String argName, Double arg) {
        sqlArgs.argDouble(argName, arg);
        return this;
    }

    
    public Sql argBigDecimal(BigDecimal arg) {
        sqlArgs.argBigDecimal(arg);
        return this;
    }

    
    public Sql argBigDecimal( String argName, BigDecimal arg) {
        sqlArgs.argBigDecimal(argName, arg);
        return this;
    }

    
    public Sql argString(String arg) {
        sqlArgs.argString(arg);
        return this;
    }

    
    public Sql argString( String argName, String arg) {
        sqlArgs.argString(argName, arg);
        return this;
    }
    
    public Sql argDate(LocalDateTime arg) {
        sqlArgs.argDate(arg);
        return this;
    }

    
    public Sql argDate( String argName, LocalDateTime arg) {
        sqlArgs.argDate(argName, arg);
        return this;
    }

    
    public Sql argDateNowPerApp() {
        sqlArgs.argDateNowPerApp();
        return this;
    }

    
    public Sql argDateNowPerApp( String argName) {
        sqlArgs.argDateNowPerApp(argName);
        return this;
    }

    
    public Sql argDateNowPerDb() {
        sqlArgs.argDateNowPerDb();
        return this;
    }

    
    public Sql argDateNowPerDb( String argName) {
        sqlArgs.argDateNowPerDb(argName);
        return this;
    }

    
    public Sql argBlobBytes(byte[] arg) {
        sqlArgs.argBlobBytes(arg);
        return this;
    }

    
    public Sql argBlobBytes( String argName, byte[] arg) {
        sqlArgs.argBlobBytes(argName, arg);
        return this;
    }

    
    public Sql argBlobInputStream(InputStream arg) {
        sqlArgs.argBlobInputStream(arg);
        return this;
    }

    
    public Sql argBlobInputStream( String argName, InputStream arg) {
        sqlArgs.argBlobInputStream(argName, arg);
        return this;
    }

    
    public Sql argClobString(String arg) {
        sqlArgs.argClobString(arg);
        return this;
    }

    
    public Sql argClobString( String argName, String arg) {
        sqlArgs.argClobString(argName, arg);
        return this;
    }

    
    public Sql argClobReader(Reader arg) {
        sqlArgs.argClobReader(arg);
        return this;
    }

    
    public Sql argClobReader( String argName, Reader arg) {
        sqlArgs.argClobReader(argName, arg);
        return this;
    }

    @Override
    public void apply(SqlSelect select) {
        if (batched != null) {
            throw new DatabaseException("Batch not supported for select");
        }
        sqlArgs.apply(select);
    }

    @Override
    public void apply(SqlInsert insert) {
        if (batched != null) {
            batch();
            for (SqlArgs args : batched) {
                args.apply(insert);
                insert.batch();
            }
        } else {
            sqlArgs.apply(insert);
        }
    }

    @Override
    public void apply(SqlUpdate update) {
        if (batched != null) {
            throw new DatabaseException("Batch not supported for update");
        }

        sqlArgs.apply(update);
    }

    public enum ColumnType {
        Integer, Long, Float, Double, BigDecimal, String, ClobString, ClobStream,
        BlobBytes, BlobStream, Date, DateNowPerApp, DateNowPerDb, Boolean
    }

    private static class Invocation {
        ColumnType columnType;
        String argName;
        Object arg;

        Invocation(ColumnType columnType, String argName, Object arg) {
            this.columnType = columnType;
            this.argName = argName;
            this.arg = arg;
        }
    }
}
