package org.xbib.jdbc.query;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;

/**
 * Interface for configuring (setting parameters) and executing a chunk of SQL.
 */
public interface SqlInsert {
    
    SqlInsert argBoolean(Boolean arg);

    SqlInsert argBoolean( String argName, Boolean arg);

    SqlInsert argInteger(Integer arg);

    SqlInsert argInteger( String argName, Integer arg);

    SqlInsert argLong(Long arg);

    SqlInsert argLong( String argName, Long arg);

    SqlInsert argFloat(Float arg);

    SqlInsert argFloat( String argName, Float arg);

    SqlInsert argDouble(Double arg);

    SqlInsert argDouble( String argName, Double arg);

    SqlInsert argBigDecimal(BigDecimal arg);

    SqlInsert argBigDecimal(String argName, BigDecimal arg);

    SqlInsert argString(String arg);

    SqlInsert argString( String argName, String arg);

    SqlInsert argDate(LocalDateTime arg); // date with time

    SqlInsert argDate(String argName, LocalDateTime arg); // date with time

    SqlInsert argLocalDate(LocalDate arg); // date only - no timestamp

    SqlInsert argLocalDate( String argName, LocalDate arg);  // date only - no timestamp

    SqlInsert argDateNowPerApp();

    SqlInsert argDateNowPerApp( String argName);

    SqlInsert argDateNowPerDb();

    SqlInsert argDateNowPerDb( String argName);

    SqlInsert argBlobBytes(byte[] arg);

    SqlInsert argBlobBytes( String argName, byte[] arg);

    SqlInsert argBlobStream(InputStream arg);

    SqlInsert argBlobStream( String argName, InputStream arg);

    SqlInsert argClobString(String arg);

    SqlInsert argClobString( String argName, String arg);

    SqlInsert argClobReader(Reader arg);

    SqlInsert argClobReader( String argName, Reader arg);

    SqlInsert withArgs(SqlArgs args);

    SqlInsert apply(Apply apply);

    /**
     * Call this between setting rows of parameters for a SQL statement. You may call it before
     * setting any parameters, after setting all, or multiple times between rows. This feature
     * only currently works with basic inserts (you can't do insertReturning type operations).
     */
    SqlInsert batch();

    /**
     * Perform the insert into the database without any verification of how many rows
     * were affected.
     *
     * @return the number of rows affected
     */
    int insert();

    /**
     * Perform the insert into the database. This will automatically verify
     * that the specified number of rows was affected, and throw a {@link WrongNumberOfRowsException}
     * if it does not match.
     */
    void insert(int expectedRowsUpdated);

    /**
     * Insert multiple rows in one database call. This will automatically verify
     * that exactly 1 row is affected for each row of parameters.
     */
    void insertBatch();

    /**
     * Insert multiple rows in one database call. This returns the results for
     * each row so you can check them yourself.
     *
     * @return an array with an element for each row in the batch; the value
     * of each array indicates how many rows were affected; note that
     * some database/driver combinations do now return this information
     * (for example, older versions of Oracle return -2 rather than the
     * number of rows)
     */
    int[] insertBatchUnchecked();

    /**
     * Use this method in conjunction with argPkSeq() to optimize inserts where the
     * primary key is being populated from a database sequence at insert time. If the
     * database can't support this feature it will be simulated with a select and then
     * the insert.
     *
     * <p>This version of insert expects exactly one row to be inserted, and will throw
     * a DatabaseException if that isn't the case.</p>
     */
    
    Long insertReturningPkSeq(String primaryKeyColumnName);

    <T> T insertReturning(String tableName, String primaryKeyColumnName, RowsHandler<T> rowsHandler,
                          String... otherColumnNames);

    SqlInsert argPkSeq( String sequenceName);

    /**
     * Use this method to populate the primary key value (assumed to be type Long)
     * from a sequence in the database. This can be used standalone, but is intended
     * to be used in conjunction with insertReturningPkSeq() to both insert and obtain
     * the inserted value in an optimized way (if possible). For databases that are
     * unable to return the value from the insert (such as Derby) this will be simulated
     * first issuing a select to read the sequence, then an insert.
     */
    SqlInsert argPkSeq( String argName,  String sequenceName);

    SqlInsert argPkLong(Long pkValue);

    SqlInsert argPkLong(String argName, Long pkValue);

    interface Apply {
        void apply(SqlInsert insert);
    }
}
