package org.xbib.jdbc.query;

import org.xbib.jdbc.query.util.DebugSql;
import org.xbib.jdbc.query.util.InternalStringReader;
import org.xbib.jdbc.query.util.Metric;
import org.xbib.jdbc.query.util.RewriteArg;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is the key class for configuring (query parameters) and executing a database query.
 */
public class SqlInsertImpl implements SqlInsert {
    
    private static final Logger log = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;
    
    private final StatementAdaptor adaptor;
    
    private final String sql;
    
    private final Options options;
    
    private List<Batch> batched;
    
    private List<Object> parameterList;       // !null ==> traditional ? args
    
    private Map<String, Object> parameterMap; // !null ==> named :abc args
    
    private String pkArgName;
    
    private int pkPos;
    
    private String pkSeqName;
    
    private Long pkLong;
    
    public SqlInsertImpl(Connection connection, String sql, Options options) {
        this.connection = connection;
        this.sql = sql;
        this.options = options;
        adaptor = new StatementAdaptor(options);
    }

    
    @Override
    public SqlInsert argBoolean(Boolean arg) {
        return positionalArg(adaptor.nullString(booleanToString(arg)));
    }

    
    @Override
    public SqlInsert argBoolean( String argName, Boolean arg) {
        return namedArg(argName, adaptor.nullString(booleanToString(arg)));
    }

    @Override
    
    public SqlInsert argInteger(Integer arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argInteger( String argName, Integer arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argLong(Long arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argLong( String argName, Long arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argFloat(Float arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argFloat( String argName, Float arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argDouble(Double arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argDouble( String argName, Double arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argBigDecimal(BigDecimal arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlInsert argBigDecimal( String argName, BigDecimal arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlInsert argString(String arg) {
        return positionalArg(adaptor.nullString(arg));
    }

    @Override
    public SqlInsert argString( String argName, String arg) {
        return namedArg(argName, adaptor.nullString(arg));
    }

    @Override
    public SqlInsert argDate(LocalDateTime arg) {
        return positionalArg(adaptor.nullDate(arg));
    }

    @Override
    public SqlInsert argDate( String argName, LocalDateTime arg) {
        return namedArg(argName, adaptor.nullDate(arg));
    }

    @Override
    public SqlInsert argLocalDate( String argName, LocalDate arg) {
        return namedArg(argName, adaptor.nullLocalDate(arg));
    }

    @Override
    public SqlInsert argLocalDate(LocalDate arg) {
        return positionalArg(adaptor.nullLocalDate(arg));
    }

    
    @Override
    public SqlInsert argDateNowPerApp() {
        return positionalArg(adaptor.nullDate(options.currentDate()));
    }

    @Override
    public SqlInsert argDateNowPerApp( String argName) {
        return namedArg(argName, adaptor.nullDate(options.currentDate()));
    }
    
    @Override
    public SqlInsert argDateNowPerDb() {
        if (options.useDatePerAppOnly()) {
            return positionalArg(adaptor.nullDate(options.currentDate()));
        }
        return positionalArg(new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    
    public SqlInsert argDateNowPerDb( String argName) {
        if (options.useDatePerAppOnly()) {
            return namedArg(argName, adaptor.nullDate(options.currentDate()));
        }
        return namedArg(argName, new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    
    public SqlInsert argBlobBytes(byte[] arg) {
        return positionalArg(adaptor.nullBytes(arg));
    }

    @Override
    
    public SqlInsert argBlobBytes( String argName, byte[] arg) {
        return namedArg(argName, adaptor.nullBytes(arg));
    }

    @Override
    
    public SqlInsert argBlobStream(InputStream arg) {
        return positionalArg(adaptor.nullInputStream(arg));
    }

    @Override
    
    public SqlInsert argBlobStream( String argName, InputStream arg) {
        return namedArg(argName, adaptor.nullInputStream(arg));
    }

    @Override
    
    public SqlInsert argClobString(String arg) {
        return positionalArg(adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    
    public SqlInsert argClobString( String argName, String arg) {
        return namedArg(argName, adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    
    public SqlInsert argClobReader(Reader arg) {
        return positionalArg(adaptor.nullClobReader(arg));
    }

    @Override
    
    public SqlInsert argClobReader( String argName, Reader arg) {
        return namedArg(argName, adaptor.nullClobReader(arg));
    }

    
    @Override
    public SqlInsert withArgs(SqlArgs args) {
        return apply(args);
    }

    
    @Override
    public SqlInsert apply(Apply apply) {
        apply.apply(this);
        return this;
    }

    @Override
    public SqlInsert batch() {
        if ((parameterList != null && !parameterList.isEmpty())
                || (parameterMap != null && !parameterMap.isEmpty())) {
            if (batched == null) {
                batched = new ArrayList<>();
            }
            batched.add(new Batch(parameterList, parameterMap));
            parameterList = new ArrayList<>();
            parameterMap = new HashMap<>();
        }
        return this;
    }

    @Override
    public int insert() {
        return updateInternal(0);
    }

    @Override
    public void insert(int expectedRowsUpdated) {
        updateInternal(expectedRowsUpdated);
    }

    @Override
    public void insertBatch() {
        int[] result = updateBatch();
        for (int r : result) {
            // Tolerate SUCCESS_NO_INFO for older versions of Oracle
            if (r != 1 && r != Statement.SUCCESS_NO_INFO) {
                throw new DatabaseException("Batch did not return the expected result: " + Arrays.toString(result));
            }
        }
    }

    @Override
    public int[] insertBatchUnchecked() {
        return updateBatch();
    }

    @Override
    public Long insertReturningPkSeq(String primaryKeyColumnName) {
        if (!hasPk()) {
            throw new DatabaseException("Call argPkSeq() before insertReturningPkSeq()");
        }

        if (options.flavor().supportsInsertReturning()) {
            return updateInternal(1, primaryKeyColumnName);
        } else {
            // Simulate by issuing a select for the next sequence value, inserting, and returning it
            Long pk = new SqlSelectImpl(connection, options.flavor().sequenceSelectNextVal(pkSeqName), options).queryLongOrNull();
            if (pk == null) {
                throw new DatabaseException("Unable to retrieve next sequence value from " + pkSeqName);
            }
            if (pkArgName != null) {
                namedArg(pkArgName, adaptor.nullNumeric(pk));
            } else {
                parameterList.set(pkPos, adaptor.nullNumeric(pk));
            }
            updateInternal(1);
            return pk;
        }
    }

    @Override
    public <T> T insertReturning(String tableName, String primaryKeyColumnName, RowsHandler<T> handler,
                                 String... otherColumnNames) {
        if (!hasPk()) {
            throw new DatabaseException("Identify a primary key with argPk*() before insertReturning()");
        }

        if (options.flavor().supportsInsertReturning()) {
            return updateInternal(1, primaryKeyColumnName, handler, otherColumnNames);
        } else if (pkSeqName != null) {
            // Simulate by issuing a select for the next sequence value, inserting, and returning it
            Long pk = new SqlSelectImpl(connection, options.flavor().sequenceSelectNextVal(pkSeqName), options)
                    .queryLongOrNull();
            if (pk == null) {
                throw new DatabaseException("Unable to retrieve next sequence value from " + pkSeqName);
            }
            if (pkArgName != null) {
                namedArg(pkArgName, adaptor.nullNumeric(pk));
            } else {
                parameterList.set(pkPos, adaptor.nullNumeric(pk));
            }
            updateInternal(1);
            StringBuilder sql = new StringBuilder();
            sql.append("select ").append(primaryKeyColumnName);
            for (String colName : otherColumnNames) {
                sql.append(", ").append(colName);
            }
            sql.append(" from ").append(tableName).append(" where ").append(primaryKeyColumnName).append("=?");
            return new SqlSelectImpl(connection, sql.toString(), options).argLong(pk).query(handler);
        } else if (pkLong != null) {
            // Insert the value, then do a select based on the primary key
            updateInternal(1);
            StringBuilder sql = new StringBuilder();
            sql.append("select ").append(primaryKeyColumnName);
            for (String colName : otherColumnNames) {
                sql.append(", ").append(colName);
            }
            sql.append(" from ").append(tableName).append(" where ").append(primaryKeyColumnName).append("=?");
            return new SqlSelectImpl(connection, sql.toString(), options).argLong(pkLong).query(handler);
        } else {
            // Should never happen if our safety checks worked
            throw new DatabaseException("Internal error");
        }
    }

    
    @Override
    public SqlInsert argPkSeq( String sequenceName) {
        if (hasPk() && batched == null) {
            throw new DatabaseException("Only call one argPk*() method");
        }
        if (hasPk() && (!pkSeqName.equals(sequenceName) || pkPos != parameterList.size())) {
            throw new DatabaseException("The argPkSeq() calls must be in the same position across batch records");
        }
        pkSeqName = sequenceName;
        SqlInsert sqlInsert = positionalArg(new RewriteArg(options.flavor().sequenceNextVal(sequenceName)));
        pkPos = parameterList.size() - 1;
        return sqlInsert;
    }

    @Override
    
    public SqlInsert argPkSeq( String argName,  String sequenceName) {
        if (hasPk() && batched == null) {
            throw new DatabaseException("Only call one argPk*() method");
        }
        if (hasPk() && !argName.equals(pkArgName)) {
            throw new DatabaseException("The primary key argument name must match across batch rows");
        }
        pkArgName = argName;
        pkSeqName = sequenceName;
        return namedArg(argName, new RewriteArg(options.flavor().sequenceNextVal(sequenceName)));
    }

    @Override
    
    public SqlInsert argPkLong(String argName, Long arg) {
        if (hasPk() && batched == null) {
            throw new DatabaseException("Only call one argPk*() method");
        }
        if (hasPk() && !argName.equals(pkArgName)) {
            throw new DatabaseException("The primary key argument name must match across batch rows");
        }
        pkArgName = argName;
        pkLong = arg;
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlInsert argPkLong(Long arg) {
        if (hasPk() && batched == null) {
            throw new DatabaseException("Only call one argPk*() method");
        }
        if (hasPk() && pkPos != parameterList.size()) {
            throw new DatabaseException("The argPkLong() calls must be in the same position across batch records");
        }
        pkLong = arg;
        SqlInsert sqlInsert = positionalArg(adaptor.nullNumeric(arg));
        pkPos = parameterList.size() - 1;
        return sqlInsert;
    }

    private boolean hasPk() {
        return pkArgName != null || pkSeqName != null || pkLong != null;
    }

    private int[] updateBatch() {
        batch();

        if (batched == null || batched.size() == 0) {
            throw new DatabaseException("Batch insert requires parameters");
        }

        PreparedStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] firstRowParameters = null;
        List<Object[]> parameters = new ArrayList<>();

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            for (Batch batch : batched) {
                MixedParameterSql mpSql = new MixedParameterSql(sql, batch.parameterList, batch.parameterMap);
                if (firstRowParameters == null) {
                    executeSql = mpSql.getSqlToExecute();
                    firstRowParameters = mpSql.getArgs();
                } else {
                    if (!executeSql.equals(mpSql.getSqlToExecute())) {
                        throw new DatabaseException("All rows in a batch must use parameters in the same way. \nSQL1: "
                                + executeSql + "\nSQL2: " + mpSql.getSqlToExecute());
                    }
                }
                parameters.add(mpSql.getArgs());
            }

            if (connection != null) {
                ps = connection.prepareStatement(executeSql);

                for (Object[] params : parameters) {
                    adaptor.addParameters(ps, params);
                    ps.addBatch();
                }

                metric.checkpoint("prep");
                int[] numAffectedRows = ps.executeBatch();
                metric.checkpoint("execBatch", parameters.size());
                isSuccess = true;
                return numAffectedRows;
            } else {
                return new int[]{-1};
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, firstRowParameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Insert", log, metric, executeSql, firstRowParameters, options);
            } else {
                DebugSql.logError("Insert", log, metric, errorCode, executeSql, firstRowParameters, options, logEx);
            }
        }
    }

    private int updateInternal(int expectedNumAffectedRows) {
        if (batched != null) {
            throw new DatabaseException("Call insertBatch() if you are using the batch() feature");
        }

        PreparedStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] parameters = null;

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();

            if (connection != null) {
                ps = connection.prepareStatement(executeSql);

                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                int numAffectedRows = ps.executeUpdate();
                metric.checkpoint("exec", numAffectedRows);
                if (expectedNumAffectedRows > 0 && numAffectedRows != expectedNumAffectedRows) {
                    errorCode = options.generateErrorCode();
                    throw new WrongNumberOfRowsException("The number of affected rows was " + numAffectedRows + ", but "
                            + expectedNumAffectedRows + " were expected." + "\n"
                            + DebugSql.exceptionMessage(executeSql, parameters, errorCode, options));
                }
                isSuccess = true;
                return numAffectedRows;
            } else {
                return -1;
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Insert", log, metric, executeSql, parameters, options);
            } else {
                DebugSql.logError("Insert", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }

    private Long updateInternal(int expectedNumAffectedRows,  String pkToReturn) {
        if (batched != null) {
            throw new DatabaseException("Call insertBatch() if you are using the batch() feature");
        }

        PreparedStatement ps = null;
        ResultSet rs = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] parameters = null;

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();

            if (connection != null) {
                ps = connection.prepareStatement(executeSql, new String[]{pkToReturn});

                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                int numAffectedRows = ps.executeUpdate();
                metric.checkpoint("exec", numAffectedRows);
                if (expectedNumAffectedRows > 0 && numAffectedRows != expectedNumAffectedRows) {
                    errorCode = options.generateErrorCode();
                    throw new WrongNumberOfRowsException("The number of affected rows was " + numAffectedRows + ", but "
                            + expectedNumAffectedRows + " were expected." + "\n"
                            + DebugSql.exceptionMessage(executeSql, parameters, errorCode, options));
                }
                rs = ps.getGeneratedKeys();
                Long pk = null;
                if (rs != null && rs.next()) {
                    pk = rs.getLong(1);
                }
                isSuccess = true;
                return pk;
            } else {
                return null;
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(rs, log);
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Insert", log, metric, executeSql, parameters, options);
            } else {
                DebugSql.logError("Insert", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }

    private <T> T updateInternal(int expectedNumAffectedRows,  String pkToReturn, RowsHandler<T> handler,
                                 String... otherCols) {
        if (batched != null) {
            throw new DatabaseException("Call insertBatch() if you are using the batch() feature");
        }

        PreparedStatement ps = null;
        ResultSet rs = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] parameters = null;

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();

            String[] returnCols = new String[otherCols.length + 1];
            returnCols[0] = pkToReturn;
            System.arraycopy(otherCols, 0, returnCols, 1, otherCols.length);

            if (connection != null) {
                ps = connection.prepareStatement(executeSql, returnCols);

                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                int numAffectedRows = ps.executeUpdate();
                metric.checkpoint("exec", numAffectedRows);
                if (expectedNumAffectedRows > 0 && numAffectedRows != expectedNumAffectedRows) {
                    errorCode = options.generateErrorCode();
                    throw new WrongNumberOfRowsException("The number of affected rows was " + numAffectedRows + ", but "
                            + expectedNumAffectedRows + " were expected." + "\n"
                            + DebugSql.exceptionMessage(executeSql, parameters, errorCode, options));
                }
                rs = ps.getGeneratedKeys();
                final ResultSet finalRs = rs;
                T result = handler.process(new RowsAdaptor(finalRs, options));
                metric.checkpoint("read");
                isSuccess = true;
                return result;
            } else {
                return null;
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(rs, log);
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Insert", log, metric, executeSql, parameters, options);
            } else {
                DebugSql.logError("Insert", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }

    private SqlInsert positionalArg(Object arg) {
        if (parameterList == null) {
            parameterList = new ArrayList<>();
        }
        parameterList.add(arg);
        return this;
    }

    private SqlInsert namedArg(String argName, Object arg) {
        if (parameterMap == null) {
            parameterMap = new HashMap<>();
        }
        if (argName.startsWith(":")) {
            argName = argName.substring(1);
        }
        parameterMap.put(argName, arg);
        return this;
    }

    private String booleanToString(Boolean b) {
        return b == null ? null : b ? "Y" : "N";
    }

    private class Batch {
        private final List<Object> parameterList;       // !null ==> traditional ? args
        private final Map<String, Object> parameterMap; // !null ==> named :abc args

        public Batch(List<Object> parameterList, Map<String, Object> parameterMap) {
            this.parameterList = parameterList;
            this.parameterMap = parameterMap;
        }
    }
}
