package org.xbib.jdbc.query;

import org.xbib.jdbc.query.util.DebugSql;
import org.xbib.jdbc.query.util.Metric;
import org.xbib.jdbc.query.util.RewriteArg;

import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is the key class for configuring (query parameters) and executing a database query.
 */
public class SqlSelectImpl implements SqlSelect {
    
    private static final Logger log = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;
    
    private final StatementAdaptor adaptor;

    private final Object cancelLock = new Object();

    private final String sql;

    private final Options options;

    private PreparedStatement ps; // hold reference to support cancel from another thread

    private List<Object> parameterList;       // !null ==> traditional ? args

    private Map<String, Object> parameterMap; // !null ==> named :abc args

    private int timeoutSeconds = -1; // -1 ==> no timeout

    private int maxRows = -1; // -1 ==> unlimited

    private int fetchSize = -1; // -1 ==> do not call setFetchSize()

    public SqlSelectImpl(Connection connection, String sql, Options options) {
        this.connection = connection;
        this.sql = sql;
        this.options = options;
        adaptor = new StatementAdaptor(options);
    }

    
    @Override
    public SqlSelect argBoolean(Boolean arg) {
        return positionalArg(adaptor.nullString(booleanToString(arg)));
    }

    
    @Override
    public SqlSelect argBoolean( String argName, Boolean arg) {
        return namedArg(argName, adaptor.nullString(booleanToString(arg)));
    }

    
    @Override
    public SqlSelect argInteger(Integer arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argInteger( String argName, Integer arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argLong(Long arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argLong( String argName, Long arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argFloat(Float arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argFloat( String argName, Float arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argDouble(Double arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argDouble( String argName, Double arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argBigDecimal(BigDecimal arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argBigDecimal( String argName, BigDecimal arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    
    @Override
    public SqlSelect argString(String arg) {
        return positionalArg(adaptor.nullString(arg));
    }

    
    @Override
    public SqlSelect argString( String argName, String arg) {
        return namedArg(argName, adaptor.nullString(arg));
    }

    
    @Override
    public SqlSelect argDate(LocalDateTime arg) {
        // Date with time
        return positionalArg(adaptor.nullDate(arg));
    }

    
    @Override
    public SqlSelect argDate( String argName, LocalDateTime arg) {
        // Date with time
        return namedArg(argName, adaptor.nullDate(arg));
    }

    
    @Override
    public SqlSelect argLocalDate(LocalDate arg) {
        // Date with no time
        return positionalArg(adaptor.nullLocalDate(arg));
    }

    
    @Override
    public SqlSelect argLocalDate( String argName, LocalDate arg) {
        // Date with no time
        return namedArg(argName, adaptor.nullLocalDate(arg));
    }

    
    @Override
    public SqlSelect argDateNowPerApp() {
        return positionalArg(adaptor.nullDate(options.currentDate()));
    }

    
    @Override
    public SqlSelect argDateNowPerApp( String argName) {
        return namedArg(argName, adaptor.nullDate(options.currentDate()));
    }

    
    @Override
    public SqlSelect argDateNowPerDb() {
        if (options.useDatePerAppOnly()) {
            return positionalArg(adaptor.nullDate(options.currentDate()));
        }
        return positionalArg(new RewriteArg(options.flavor().dbTimeMillis()));
    }

    
    @Override
    public SqlSelect argDateNowPerDb( String argName) {
        if (options.useDatePerAppOnly()) {
            return namedArg(argName, adaptor.nullDate(options.currentDate()));
        }
        return namedArg(argName, new RewriteArg(options.flavor().dbTimeMillis()));
    }

    
    @Override
    public SqlSelect withTimeoutSeconds(int seconds) {
        timeoutSeconds = seconds;
        return this;
    }

    
    @Override
    public SqlSelect withMaxRows(int rows) {
        maxRows = rows;
        return this;
    }

    
    @Override
    public SqlSelect withArgs(SqlArgs args) {
        return apply(args);
    }

    
    @Override
    public SqlSelect apply(Apply apply) {
        apply.apply(this);
        return this;
    }

    
    @Override
    public SqlSelect fetchSize(int rows) {
        fetchSize = rows;
        return this;
    }

    @Override
    public Boolean queryBooleanOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getBooleanOrNull();
            }
            return null;
        });
    }

    @Override
    public boolean queryBooleanOrFalse() {
        Boolean result = queryBooleanOrNull();
        return result != null && result;
    }

    @Override
    public boolean queryBooleanOrTrue() {
        Boolean result = queryBooleanOrNull();
        return result == null || result;
    }

    @Override
    
    public Long queryLongOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getLongOrNull(1);
            }
            return null;
        });
    }

    @Override
    public long queryLongOrZero() {
        Long l = queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getLongOrZero(1);
            }
            return 0L;
        });
        return l != null ? l : 0L;
    }
    
    @Override
    public List<Long> queryLongs() {
        return queryWithTimeout(rs -> {
            List<Long> result = new ArrayList<>();
            while (rs.next()) {
                Long value = rs.getLongOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }

    
    @Override
    public Integer queryIntegerOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getIntegerOrNull(1);
            }
            return null;
        });
    }

    @Override
    public int queryIntegerOrZero() {
        Integer i = queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getIntegerOrZero(1);
            }
            return 0;
        });
        return i != null ? i : 0;
    }
    
    @Override
    public List<Integer> queryIntegers() {
        return queryWithTimeout(rs -> {
            List<Integer> result = new ArrayList<>();
            while (rs.next()) {
                Integer value = rs.getIntegerOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }

    
    @Override
    public Float queryFloatOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getFloatOrNull(1);
            }
            return null;
        });
    }

    @Override
    public float queryFloatOrZero() {
        Float f = queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getFloatOrZero(1);
            }
            return 0f;
        });
        return f!= null ? f : 0f;
    }
    
    @Override
    public List<Float> queryFloats() {
        return queryWithTimeout(rs -> {
            List<Float> result = new ArrayList<>();
            while (rs.next()) {
                Float value = rs.getFloatOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }
    
    @Override
    public Double queryDoubleOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getDoubleOrNull(1);
            }
            return null;
        });
    }

    @Override
    public double queryDoubleOrZero() {
        Double d = queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getDoubleOrZero(1);
            }
            return 0d;
        });
        return d != null ? d : 0d;
    }
    
    @Override
    public List<Double> queryDoubles() {
        return queryWithTimeout(rs -> {
            List<Double> result = new ArrayList<>();
            while (rs.next()) {
                Double value = rs.getDoubleOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }
    
    @Override
    public BigDecimal queryBigDecimalOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getBigDecimalOrNull(1);
            }
            return null;
        });
    }

    
    @Override
    public BigDecimal queryBigDecimalOrZero() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getBigDecimalOrZero(1);
            }
            return new BigDecimal(0);
        });
    }

    
    @Override
    public List<BigDecimal> queryBigDecimals() {
        return queryWithTimeout(rs -> {
            List<BigDecimal> result = new ArrayList<>();
            while (rs.next()) {
                BigDecimal value = rs.getBigDecimalOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }

    @Override
    public String queryStringOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getStringOrNull(1);
            }
            return null;
        });
    }

    
    @Override
    public String queryStringOrEmpty() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getStringOrEmpty(1);
            }
            return "";
        });
    }
    
    @Override
    public List<String> queryStrings() {
        return queryWithTimeout(rs -> {
            List<String> result = new ArrayList<>();
            while (rs.next()) {
                String value = rs.getStringOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }
    
    @Override
    public LocalDateTime queryDateOrNull() {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getLocalDateTimeOrNull(1);
            }
            return null;
        });
    }
    
    @Override
    public List<LocalDateTime> queryDates() {
        return queryWithTimeout(rs -> {
            List<LocalDateTime> result = new ArrayList<>();
            while (rs.next()) {
                LocalDateTime value = rs.getLocalDateTimeOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }
    
    @Override
    public LocalDate queryLocalDateOrNull() {
        // Date without time
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rs.getLocalDateOrNull(1);
            }
            return null;
        });
    }
    
    @Override
    public List<LocalDate> queryLocalDates() {
        // Date without time
        return queryWithTimeout(rs -> {
            List<LocalDate> result = new ArrayList<>();
            while (rs.next()) {
                LocalDate value = rs.getLocalDateOrNull(1);
                if (value != null) {
                    result.add(value);
                }
            }
            return result;
        });
    }

    @Override
    public <T> T query(RowsHandler<T> rowsHandler) {
        return queryWithTimeout(rowsHandler);
    }

    @Override
    public <T> T queryOneOrNull(final RowHandler<T> rowHandler) {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                T result = rowHandler.process(rs);
                if (rs.next()) {
                    throw new ConstraintViolationException("Expected exactly one row to be returned but found multiple");
                }
                return result;
            }
            return null;
        });
    }

    @Override
    public <T> T queryOneOrThrow(RowHandler<T> rowHandler) {
        T result = queryOneOrNull(rowHandler);
        if (result == null) {
            throw new ConstraintViolationException("Expected exactly one row to be returned but found none");
        }
        return result;
    }

    @Override
    public <T> T queryFirstOrNull(final RowHandler<T> rowHandler) {
        return queryWithTimeout(rs -> {
            if (rs.next()) {
                return rowHandler.process(rs);
            }
            return null;
        });
    }

    @Override
    public <T> T queryFirstOrThrow(RowHandler<T> rowHandler) {
        T result = queryFirstOrNull(rowHandler);
        if (result == null) {
            throw new ConstraintViolationException("Expected one or more rows to be returned but found none");
        }
        return result;
    }

    @Override
    public <T> List<T> queryMany(final RowHandler<T> rowHandler) {
        return queryWithTimeout(rs -> {
            List<T> result = new ArrayList<>();
            while (rs.next()) {
                T row = rowHandler.process(rs);
                if (row != null) {
                    result.add(row);
                }
            }
            return result;
        });
    }

    private SqlSelect positionalArg(Object arg) {
        if (parameterList == null) {
            parameterList = new ArrayList<>();
        }
        parameterList.add(arg);
        return this;
    }

    private SqlSelect namedArg(String argName, Object arg) {
        if (parameterMap == null) {
            parameterMap = new HashMap<>();
        }
        if (argName.startsWith(":")) {
            argName = argName.substring(1);
        }
        parameterMap.put(argName, arg);
        return this;
    }

    private String booleanToString(Boolean b) {
        return b == null ? null : b ? "Y" : "N";
    }

    private <T> T queryWithTimeout(RowsHandler<T> handler) {
        assert ps == null;
        ResultSet rs = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] parameters = null;

        boolean isWarn = false;
        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();

            if (connection != null) {
                synchronized (cancelLock) {
                    ps = connection.prepareStatement(executeSql,
                            ResultSet.TYPE_SCROLL_INSENSITIVE,
                            ResultSet.CONCUR_READ_ONLY);
                }

                if (timeoutSeconds >= 0) {
                    ps.setQueryTimeout(timeoutSeconds);
                }

                if (maxRows > 0) {
                    ps.setMaxRows(maxRows);
                }

                if (fetchSize >= 0) {
                    ps.setFetchSize(fetchSize);
                }

                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                rs = ps.executeQuery();
                metric.checkpoint("exec");
                final ResultSet finalRs = rs;
                T result = handler.process(new RowsAdaptor(finalRs, options));
                metric.checkpoint("read");
                isSuccess = true;
                return result;
            } else {
                return null;
            }
        } catch (SQLException e) {
            if (e.getErrorCode() == 1013) {
                isWarn = true;
                // It's ambiguous based on the Oracle error code whether it was a timeout or cancel
                throw new QueryTimedOutException("Timeout of " + timeoutSeconds + " seconds exceeded or user cancelled", e);
            }
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(rs, log);
            adaptor.closeQuietly(ps, log);
            synchronized (cancelLock) {
                ps = null;
            }
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Query", log, metric, executeSql, parameters, options);
            } else if (isWarn) {
                DebugSql.logWarning("Query", log, metric, "QueryTimedOutException", executeSql, parameters, options, null);
            } else {
                DebugSql.logError("Query", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }
}
