package org.xbib.jdbc.query;

import org.xbib.jdbc.query.util.DebugSql;
import org.xbib.jdbc.query.util.InternalStringReader;
import org.xbib.jdbc.query.util.Metric;
import org.xbib.jdbc.query.util.RewriteArg;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is the key class for configuring (query parameters) and executing a database query.
 */
public class SqlUpdateImpl implements SqlUpdate {
    
    private static final Logger log = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;
    
    private final StatementAdaptor adaptor;
    
    private final String sql;
    
    private final Options options;
    
    private List<Object> parameterList;
    
    private Map<String, Object> parameterMap;
    
    public SqlUpdateImpl(Connection connection, String sql, Options options) {
        this.connection = connection;
        this.sql = sql;
        this.options = options;
        adaptor = new StatementAdaptor(options);
    }

    
    @Override
    public SqlUpdate argBoolean(Boolean arg) {
        return positionalArg(adaptor.nullString(booleanToString(arg)));
    }

    
    @Override
    public SqlUpdate argBoolean(String argName, Boolean arg) {
        return namedArg(argName, adaptor.nullString(booleanToString(arg)));
    }

    @Override
    
    public SqlUpdate argInteger(Integer arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argInteger(String argName, Integer arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argLong(Long arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argLong(String argName, Long arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argFloat(Float arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argFloat(String argName, Float arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argDouble(Double arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argDouble(String argName, Double arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argBigDecimal(BigDecimal arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argBigDecimal(String argName, BigDecimal arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    
    public SqlUpdate argString(String arg) {
        return positionalArg(adaptor.nullString(arg));
    }

    @Override
    
    public SqlUpdate argString(String argName, String arg) {
        return namedArg(argName, adaptor.nullString(arg));
    }

    @Override
    
    public SqlUpdate argDate(LocalDateTime arg) {
        return positionalArg(adaptor.nullDate(arg));
    }

    @Override
    
    public SqlUpdate argDate(String argName, LocalDateTime arg) {
        return namedArg(argName, adaptor.nullDate(arg));
    }

    @Override
    
    public SqlUpdate argLocalDate(LocalDate arg) {
        return positionalArg(adaptor.nullLocalDate(arg));
    }

    @Override
    
    public SqlUpdate argLocalDate(String argName, LocalDate arg) {
        return namedArg(argName, adaptor.nullLocalDate(arg));
    }
    
    @Override
    public SqlUpdate argDateNowPerApp() {
        return positionalArg(adaptor.nullDate(options.currentDate()));
    }

    @Override
    
    public SqlUpdate argDateNowPerApp(String argName) {
        return namedArg(argName, adaptor.nullDate(options.currentDate()));
    }

    @Override
    public SqlUpdate argDateNowPerDb() {
        if (options.useDatePerAppOnly()) {
            return positionalArg(adaptor.nullDate(options.currentDate()));
        }
        return positionalArg(new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    
    public SqlUpdate argDateNowPerDb(String argName) {
        if (options.useDatePerAppOnly()) {
            return namedArg(argName, adaptor.nullDate(options.currentDate()));
        }
        return namedArg(argName, new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    
    public SqlUpdate argBlobBytes(byte[] arg) {
        return positionalArg(adaptor.nullBytes(arg));
    }

    @Override
    
    public SqlUpdate argBlobBytes(String argName, byte[] arg) {
        return namedArg(argName, adaptor.nullBytes(arg));
    }

    @Override
    
    public SqlUpdate argBlobStream(InputStream arg) {
        return positionalArg(adaptor.nullInputStream(arg));
    }

    @Override
    
    public SqlUpdate argBlobStream(String argName, InputStream arg) {
        return namedArg(argName, adaptor.nullInputStream(arg));
    }

    @Override
    
    public SqlUpdate argClobString(String arg) {
        return positionalArg(adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    
    public SqlUpdate argClobString(String argName, String arg) {
        return namedArg(argName, adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    
    public SqlUpdate argClobReader(Reader arg) {
        return positionalArg(adaptor.nullClobReader(arg));
    }

    @Override
    
    public SqlUpdate argClobReader(String argName, Reader arg) {
        return namedArg(argName, adaptor.nullClobReader(arg));
    }

    
    @Override
    public SqlUpdate withArgs(SqlArgs args) {
        return apply(args);
    }

    
    @Override
    public SqlUpdate apply(Apply apply) {
        apply.apply(this);
        return this;
    }

    @Override
    public int update() {
        return updateInternal(0);
    }

    @Override
    public void update(int expectedNumAffectedRows) {
        updateInternal(expectedNumAffectedRows);
    }

    private int updateInternal(int expectedNumAffectedRows) {
        PreparedStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));

        String executeSql = sql;
        Object[] parameters = null;

        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();

            if (connection != null) {
                ps = connection.prepareStatement(executeSql);
                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                int numAffectedRows = ps.executeUpdate();
                metric.checkpoint("exec", numAffectedRows);
                if (expectedNumAffectedRows > 0 && numAffectedRows != expectedNumAffectedRows) {
                    errorCode = options.generateErrorCode();
                    throw new WrongNumberOfRowsException("The number of affected rows was " + numAffectedRows + ", but "
                            + expectedNumAffectedRows + " were expected." + "\n"
                            + DebugSql.exceptionMessage(executeSql, parameters, errorCode, options));
                }
                isSuccess = true;
                return numAffectedRows;
            } else {
                return -1;
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Update", log, metric, executeSql, parameters, options);
            } else {
                DebugSql.logError("Update", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }

    
    private SqlUpdate positionalArg(Object arg) {
        if (parameterList == null) {
            parameterList = new ArrayList<>();
        }
        parameterList.add(arg);
        return this;
    }

    
    private SqlUpdate namedArg(String argName, Object arg) {
        if (parameterMap == null) {
            parameterMap = new HashMap<>();
        }
        if (argName.startsWith(":")) {
            argName = argName.substring(1);
        }
        parameterMap.put(argName, arg);
        return this;
    }

    private String booleanToString(Boolean b) {
        return b == null ? null : b ? "Y" : "N";
    }
}
