package org.xbib.jdbc.query.test;

import org.junit.jupiter.api.Test;
import org.xbib.jdbc.query.Config;
import org.xbib.jdbc.query.ConfigFrom;

import java.io.File;
import java.io.FileWriter;
import java.math.BigDecimal;
import java.util.Properties;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Unit tests for the configuration classes.
 */
public class ConfigTest {

    @Test
    public void testSystemProperties() {
        System.setProperty("foo", "bar");
        Config config = ConfigFrom.firstOf().systemProperties().get();
        assertEquals("bar", config.getString("foo"));
        assertNull(config.getString("unknown"));
        assertEquals("default", config.getString("unknown", "default"));
    }

    @Test
    public void testProperties() {
        Properties properties = new Properties();
        properties.setProperty("foo", "bar");

        Config config = ConfigFrom.firstOf().properties(properties).get();

        assertEquals("bar", config.getString("foo"));
        assertNull(config.getString("unknown"));
        assertEquals("default", config.getString("unknown", "default"));
    }

    @Test
    public void testPropertyFiles() throws Exception {
        Properties properties = new Properties();

        properties.setProperty("foo", "1");
        String filename1 = "build/ConfigTest-properties-1.properties";
        properties.store(new FileWriter(filename1), null);

        properties.setProperty("foo", "2");
        properties.setProperty("foo2", "-2");
        String filename2 = "build/ConfigTest-properties-2.properties";
        properties.store(new FileWriter(filename2), null);

        // Throw a null in here just to make sure it doesn't blow up
        Config config = ConfigFrom.firstOf().propertyFile(filename1, null, filename2).get();

        assertEquals(Integer.valueOf(1), config.getInteger("foo"));
        assertEquals(Integer.valueOf(-2), config.getInteger("foo2"));
        assertNull(config.getInteger("unknown"));
        assertEquals(5, config.getInteger("unknown", 5));

        // Now flip the order and verify precedence works
        config = ConfigFrom.firstOf().propertyFile(filename2, null, filename1).get();
        assertEquals(Integer.valueOf(2), config.getInteger("foo"));
        assertEquals(Integer.valueOf(-2), config.getInteger("foo2"));

        // Same as above tests, but using File version rather than filename String
        config = ConfigFrom.firstOf().propertyFile(new File(filename1), new File("does not exist"), new File(filename2)).get();

        assertEquals(Integer.valueOf(1), config.getInteger("foo"));
        assertEquals(Integer.valueOf(-2), config.getInteger("foo2"));
        assertNull(config.getInteger("unknown"));
        assertEquals(5, config.getInteger("unknown", 5));

        // Now flip the order and verify precedence works
        config = ConfigFrom.firstOf().propertyFile(new File(filename2), null, new File(filename1)).get();
        assertEquals(Integer.valueOf(2), config.getInteger("foo"));
        assertEquals(Integer.valueOf(-2), config.getInteger("foo2"));
    }

    @Test
    public void testNested() {
        Config config = ConfigFrom.firstOf()
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "a" : null))
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "b" : null)).get();

        assertEquals("a", config.getString("foo"));

        // Re-mapping prefix in nested config
        config = ConfigFrom.firstOf()
                .config(ConfigFrom.firstOf().custom(key -> key.equals("a.foo") ? "a" : null).removePrefix("a."))
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "b" : null)).get();

        assertEquals("a", config.getString("foo"));

        // Excluding nested config, should skip to next
        config = ConfigFrom.firstOf()
                .config(ConfigFrom.firstOf().custom(key -> key.equals("a.foo") ? "a" : null).removePrefix("a.").excludeRegex("fo{2}"))
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "b" : null)).get();

        assertEquals("b", config.getString("foo"));
        assertNull(config.getString("foooo"));

        config = ConfigFrom.firstOf()
                .config(ConfigFrom.firstOf().custom(key -> key.equals("a.foo") ? "a" : null).excludePrefix("a.", "other."))
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "b" : null).addPrefix("a.")).get();

        assertEquals("b", config.getString("a.foo"));
        assertNull(config.getString("foo"));

        config = ConfigFrom.firstOf()
                .config(ConfigFrom.firstOf().custom(key -> key.equals("a.foo") ? "a" : null).includePrefix("other."))
                .config(ConfigFrom.firstOf().custom(key -> key.equals("foo") ? "b" : null).addPrefix("a.").includeRegex("a.*f.*")).get();

        assertEquals("b", config.getString("a.foo"));
        assertNull(config.getString("foo"));
        assertNull(config.getString("other.foo"));
    }

    @Test
    public void testStripPrefixConflict() {
        Config config = ConfigFrom.firstOf().value("a.foo", "a").value("foo", "bar").removePrefix("a.").get();

        assertEquals("bar", config.getString("foo"));
    }

    @Test
    public void testException() {
        Config config = ConfigFrom.firstOf().custom(key -> {
            throw new SecurityException("Pretending security policy is in place");
        }).get();

        // We do this call twice, but you should see only one warning in the log
        assertEquals("default", config.getString("foo", "default"));
        assertEquals("default", config.getString("foo", "default"));
    }

    @Test
    public void testTidyValues() {
        Config config = ConfigFrom.firstOf().value("foo", " a ").get();

        // Strip whitespace
        assertEquals("a", config.getString("foo"));

        config = ConfigFrom.firstOf().value("foo", " ").value("foo", "").value("foo", null).value("foo", "a").get();

        // Skip over the garbage ones
        assertEquals("a", config.getString("foo"));
    }

    @Test
    public void testBoolean() {
        // Case insensitive, allow either true/false or yes/no
        Config config = ConfigFrom.firstOf().value("foo", "tRuE").get();

        assertTrue(config.getBooleanOrFalse("foo"));
        assertTrue(config.getBooleanOrTrue("foo"));
        assertFalse(config.getBooleanOrFalse("unknown"));
        assertTrue(config.getBooleanOrTrue("unknown"));

        config = ConfigFrom.firstOf().value("foo", "yEs").get();

        assertTrue(config.getBooleanOrFalse("foo"));
        assertTrue(config.getBooleanOrTrue("foo"));
        assertFalse(config.getBooleanOrFalse("unknown"));
        assertTrue(config.getBooleanOrTrue("unknown"));

        config = ConfigFrom.firstOf().value("foo", "fAlSe").get();

        assertFalse(config.getBooleanOrFalse("foo"));
        assertFalse(config.getBooleanOrTrue("foo"));
        assertFalse(config.getBooleanOrFalse("unknown"));
        assertTrue(config.getBooleanOrTrue("unknown"));

        config = ConfigFrom.firstOf().value("foo", "nO").get();

        assertFalse(config.getBooleanOrFalse("foo"));
        assertFalse(config.getBooleanOrTrue("foo"));
        assertFalse(config.getBooleanOrFalse("unknown"));
        assertTrue(config.getBooleanOrTrue("unknown"));

        config = ConfigFrom.firstOf().value("foo", "bad value").get();

        assertFalse(config.getBooleanOrFalse("foo"));
        assertTrue(config.getBooleanOrTrue("foo"));
        assertFalse(config.getBooleanOrFalse("unknown"));
        assertTrue(config.getBooleanOrTrue("unknown"));
    }

    @Test
    public void testInteger() {
        Config config = ConfigFrom.firstOf().value("good", "123").value("bad", "hi").get();
        assertEquals(Integer.valueOf(123), config.getInteger("good"));
        assertNull(config.getInteger("bad"));
        assertNull(config.getInteger("missing"));
        assertEquals(123, config.getInteger("good", 5));
        assertEquals(5, config.getInteger("bad", 5));
        assertEquals(5, config.getInteger("missing", 5));
    }

    @Test
    public void testLong() {
        Config config = ConfigFrom.firstOf().value("good", "123").value("bad", "hi").get();

        assertEquals(Long.valueOf(123), config.getLong("good"));
        assertNull(config.getLong("bad"));
        assertNull(config.getLong("missing"));
        assertEquals(123, config.getLong("good", 5));
        assertEquals(5, config.getLong("bad", 5));
        assertEquals(5, config.getLong("missing", 5));
    }

    @Test
    public void testFloat() {
        Config config = ConfigFrom.firstOf().value("good", "123.45").value("bad", "hi").get();

        assertEquals(Float.valueOf(123.45f), config.getFloat("good"));
        assertNull(config.getFloat("bad"));
        assertNull(config.getFloat("missing"));
        assertEquals(123.45, config.getFloat("good", 5.45f), 0.001);
        assertEquals(5.45, config.getFloat("bad", 5.45f), 0.001);
        assertEquals(5.45, config.getFloat("missing", 5.45f), 0.001);
    }

    @Test
    public void testDouble() {
        Config config = ConfigFrom.firstOf().value("good", "123.45").value("bad", "hi").get();

        assertEquals(Double.valueOf(123.45), config.getDouble("good"));
        assertNull(config.getDouble("bad"));
        assertNull(config.getDouble("missing"));
        assertEquals(123.45, config.getDouble("good", 5.45), 0.001);
        assertEquals(5.45, config.getDouble("bad", 5.45), 0.001);
        assertEquals(5.45, config.getDouble("missing", 5.45), 0.001);
    }

    @Test
    public void testBigDecimal() {
        Config config = ConfigFrom.firstOf().value("good", "123.45").value("bad", "hi").get();

        assertEquals(new BigDecimal("123.45"), config.getBigDecimal("good"));
        assertNull(config.getBigDecimal("bad"));
        assertNull(config.getBigDecimal("missing"));
        assertEquals(new BigDecimal("123.45"), config.getBigDecimal("good", new BigDecimal("5.45")));
        assertEquals(new BigDecimal("5.45"), config.getBigDecimal("bad", new BigDecimal("5.45")));
        assertEquals(new BigDecimal("5.45"), config.getBigDecimal("missing", new BigDecimal("5.45")));
    }
}
