package org.xbib.jdbc.query.test;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.xbib.jdbc.query.Config;
import org.xbib.jdbc.query.ConfigFrom;
import org.xbib.jdbc.query.DatabaseProvider;
import org.xbib.jdbc.query.OptionsOverride;
import org.xbib.jdbc.query.Schema;
import org.xbib.jdbc.query.Sql;
import org.xbib.jdbc.query.SqlArgs;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;

/**
 * Exercise database functionality with a real HyperSQL database.
 */
@Disabled
public class HsqldbTest extends CommonTest {

    private static final Logger logger = Logger.getLogger(HsqldbTest.class.getName());

    @Override
    protected DatabaseProvider createDatabaseProvider(OptionsOverride options) throws Exception {
        String propertiesFile = System.getProperty("local.properties", "local.properties");
        Config config = ConfigFrom.firstOf()
                .systemProperties()
                .propertyFile(propertiesFile)
                .excludePrefix("database.")
                .removePrefix("hsqldb.").get();
        logger.log(Level.INFO, "config = " + config);
        return DatabaseProvider.fromDriverManager(config)
                .withSqlParameterLogging()
                .withSqlInExceptionMessages()
                .withOptions(options).create();
    }

    @Test
    public void noDatabaseAccess() throws Exception {
        DatabaseProvider provider = createDatabaseProvider(new OptionsOverride());
        provider.transact(dbp -> {
            // Do nothing, just making sure no exception is thrown
        });
        provider.transact((dbp, tx) -> {
            // Do nothing, just making sure no exception is thrown
        });
        provider.transact((dbp, tx) -> {
            tx.setRollbackOnError(true);
            // Do nothing, just making sure no exception is thrown
        });
        provider.transact((dbp, tx) -> {
            tx.setRollbackOnly(true);
            // Do nothing, just making sure no exception is thrown
        });
    }

    @Disabled("LocalDate implementations should be TimeZone agnostic, but HSQLDB implementation has a bug.")
    @Test
    public void argLocalDateTimeZones() {
        // See bug: https://bugs.documentfoundation.org/show_bug.cgi?id=63566
        super.argLocalDateTimeZones();
    }

    /**
     * This one is adjusted in that the float values are passed as double, because
     * the database stores them both as double and there doesn't appear to be a way
     * to tell that one was actually declared as a float.
     */
    @Test
    public void saveResultAsTable() {
        new Schema().addTable("dbtest")
                .addColumn("nbr_integer").asInteger().primaryKey().table()
                .addColumn("nbr_long").asLong().table()
                .addColumn("nbr_float").asFloat().table()
                .addColumn("nbr_double").asDouble().table()
                .addColumn("nbr_big_decimal").asBigDecimal(19, 9).table()
                .addColumn("str_varchar").asString(80).table()
                .addColumn("str_fixed").asStringFixed(1).table()
                .addColumn("str_lob").asClob().table()
                .addColumn("bin_blob").asBlob().table()
                .addColumn("boolean_flag").asBoolean().table()
                .addColumn("date_millis").asDate().table()
                .addColumn("local_date").asLocalDate().schema().execute(db);

        db.toInsert("insert into dbtest (nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal, str_varchar,"
                        + " str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date) values (?,?,?,?,?,?,?,?,?,?,?,?)")
                .argInteger(Integer.MAX_VALUE).argLong(Long.MAX_VALUE).argDouble((double) Float.MAX_VALUE)
                .argDouble(Double.MAX_VALUE).argBigDecimal(new BigDecimal("123.456"))
                .argString("hello").argString("Z").argClobString("hello again")
                .argBlobBytes(new byte[]{'1', '2'}).argBoolean(true)
                .argDateNowPerApp().argLocalDate(localDateNow).insert(1);

        db.toInsert("insert into dbtest (nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal, str_varchar,"
                        + " str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date) values (?,?,?,?,?,?,?,?,?,?,?,?)")
                .argInteger(Integer.MIN_VALUE).argLong(Long.MIN_VALUE).argDouble(0.000001d)
                .argDouble(Double.MIN_VALUE).argBigDecimal(new BigDecimal("-123.456"))
                .argString("goodbye").argString("A").argClobString("bye again")
                .argBlobBytes(new byte[]{'3', '4'}).argBoolean(false)
                .argDateNowPerApp().argLocalDate(localDateNow).insert(1);

        String expectedSchema = new Schema().addTable("dbtest2")
                .addColumn("nbr_integer").asInteger().table()
                .addColumn("nbr_long").asLong().table()
                .addColumn("nbr_float").asFloat().table()
                .addColumn("nbr_double").asDouble().table()
                .addColumn("nbr_big_decimal").asBigDecimal(19, 9).table()
                .addColumn("str_varchar").asString(80).table()
                .addColumn("str_fixed").asStringFixed(1).table()
                .addColumn("str_lob").asClob().table()
                .addColumn("bin_blob").asBlob().table()
                .addColumn("boolean_flag").asBoolean().table()
                .addColumn("date_millis").asDate().table()
                .addColumn("local_date").asLocalDate().schema().print(db.flavor());

        List<SqlArgs> args = db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                        + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest")
                .query(rs -> {
                    List<SqlArgs> result = new ArrayList<>();
                    while (rs.next()) {
                        if (result.size() == 0) {
                            db.dropTableQuietly("dbtest2");
                            Schema schema = new Schema().addTableFromRow("dbtest2", rs).schema();
                            assertEquals(expectedSchema, schema.print(db.flavor()));
                            schema.execute(db);
                        }
                        result.add(SqlArgs.readRow(rs));
                    }
                    return result;
                });

        db.toInsert(Sql.insert("dbtest2", args)).insertBatch();

        assertEquals(2, db.toSelect("select count(*) from dbtest2").queryIntegerOrZero());
        assertEquals(db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest order by 1")
                        .queryMany(SqlArgs::readRow),
                db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest2 order by 1")
                        .queryMany(SqlArgs::readRow));
        assertEquals(Arrays.asList(
                        new SqlArgs()
                                .argInteger("nbr_integer", Integer.MIN_VALUE)
                                .argLong("nbr_long", Long.MIN_VALUE)
                                .argDouble("nbr_float", 0.000001d)
                                .argDouble("nbr_double", Double.MIN_VALUE)
                                .argBigDecimal("nbr_big_decimal", new BigDecimal("-123.456"))
                                .argString("str_varchar", "goodbye")
                                .argString("str_fixed", "A")
                                .argClobString("str_lob", "bye again")
                                .argBlobBytes("bin_blob", new byte[]{'3', '4'})
                                .argString("boolean_flag", "N")//.argBoolean("boolean_flag", false)
                                .argDate("date_millis", now)
                                .argLocalDate("local_date", localDateNow),
                        new SqlArgs()
                                .argInteger("nbr_integer", Integer.MAX_VALUE)
                                .argLong("nbr_long", Long.MAX_VALUE)
                                .argDouble("nbr_float", (double) Float.MAX_VALUE)
                                .argDouble("nbr_double", Double.MAX_VALUE)
                                .argBigDecimal("nbr_big_decimal", new BigDecimal("123.456"))
                                .argString("str_varchar", "hello")
                                .argString("str_fixed", "Z")
                                .argClobString("str_lob", "hello again")
                                .argBlobBytes("bin_blob", new byte[]{'1', '2'})
                                .argString("boolean_flag", "Y")//.argBoolean("boolean_flag", true)
                                .argDate("date_millis", now)
                                .argLocalDate("local_date", localDateNow)),
                db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest2 order by 1")
                        .queryMany(SqlArgs::readRow));
    }
}
